import asyncio
import signal

import configargparse
from structlog import get_logger

from alarme import Application
from alarme.scripts.common import init_logging, uncaught_exception, loop_uncaught_exception


def exit_handler(app, logger, sig):
    logger.info('application_signal', name=sig.name, value=sig.value)
    app.stop()


def run(config_path, log,
        core_application_factory=Application,
        ):
    # Logging init
    logger = get_logger()
    loop = asyncio.get_event_loop()
    init_logging(log, 'server')
    loop.set_exception_handler(loop_uncaught_exception)

    # Core init
    logger.info('application_init')
    core_app = core_application_factory(exception_handler=uncaught_exception)
    loop.run_until_complete(core_app.load_config(config_path))

    for signal_code in (signal.SIGINT, signal.SIGTERM):
        loop.add_signal_handler(signal_code, exit_handler, core_app, logger, signal_code)

    loop.run_until_complete(core_app.run())
    loop.close()


def main():
    parser = configargparse.ArgParser(description='Alarm system software for Raspberry Pi')
    parser.add_argument('-gc', '--generic-config', is_config_file=True, help='Generic config')
    parser.add_argument('-c', '--config', help='Config directory')
    parser.add_argument('-l', '--log', type=str, default='/var/log/alarme', help='Logs dir')
    args = parser.parse_args()
    run(args.config, args.log)


if __name__ == '__main__':
    main()



from .gsm_action import GsmAction


class CallAction(GsmAction):

    def __init__(self, app, id_, serial_url, number, seconds):
        super().__init__(app, id_, serial_url)
        self.number = number
        self.seconds = seconds

    async def run(self):
        self.logger.info('make_call')
        await self.gsm_modem.call(self.number, self.seconds)



from aiohttp.web import HTTPFound

from .core import CoreView
from ..util import login_required, handle_exception


class Home(CoreView):

    @login_required
    async def req(self):
        return HTTPFound(self.request.app.router.get('control').url())

    @handle_exception
    async def get(self):
        return await self.req()

    @handle_exception
    async def post(self):
        return await self.req()



import json

from aiohttp import BasicAuth
from aiohttp.web import HTTPUnauthorized, Response


AUTH_REQUIRED_HEADERS = {'WWW-Authenticate': 'Basic realm="User Visible Realm"'}
STATE_COLOR_PREFIX = '@'


def login_required(func):
    async def wrapped(self, *args, **kwargs):
        if (self.sensor.login, self.sensor.password) == (None, None):
            return await func(self, *args, **kwargs)
        auth_header = self.request.headers.get('Authorization')
        if auth_header:
            auth = BasicAuth.decode(auth_header)
            if (auth.login, auth.password) == (self.sensor.login, self.sensor.password):
                return await func(self, *args, **kwargs)
        return HTTPUnauthorized(headers=AUTH_REQUIRED_HEADERS)
    return wrapped


def handle_exception(func):
    async def wrapped(self, *args, **kwargs):
        try:
            return await func(self, *args, **kwargs)
        except Exception as err:
            self.sensor.app.exception_handler(err)
            raise
    return wrapped


def json_response(func):
    async def wrapped(self, *args, **kwargs):
        return Response(text=json.dumps(await func(self, *args, **kwargs)), content_type='text/json')
    return wrapped


def expand_color(color, states):
    if color.strip().startswith(STATE_COLOR_PREFIX):
        return states[color.replace(STATE_COLOR_PREFIX, '')].color
    return color



import asyncio

from alarme import Action
from alarme.extras.common import SingleRFDevice


class RfTransmitterAction(Action):

    def __init__(self, app, id_, gpio, code, code_extra=0, run_count=1, run_interval=0.02):
        super().__init__(app, id_)
        self.gpio = gpio
        self.code = code
        self.code_extra = code_extra
        self.run_count = run_count
        self.run_interval = run_interval
        self.rf_device = SingleRFDevice(self.gpio)

    def _continue(self, run_count):
        return self.running and (self.run_count is None or run_count < self.run_count)

    async def run(self):
        self.rf_device.enable_tx()
        try:
            run_count = 0
            while self._continue(run_count):
                self.rf_device.tx_code(self.code + self.code_extra)
                run_count += 1
                if self._continue(run_count):
                    await asyncio.sleep(self.run_interval)
        finally:
            self.rf_device.disable_tx()

    async def cleanup(self):
        await super().cleanup()
        # self.rf_device.cleanup()



import asyncio
import os.path
from importlib import import_module

import yaml
from structlog import get_logger

from .state import State
from .action_descriptor import ActionDescriptor
from .schedule import Schedule


MAIN_CONFIG_FILE = 'config.yaml'


def import_class(class_str):
    module_name, class_name = class_str.rsplit('.', 1)
    module = import_module(module_name)
    return getattr(module, class_name)


def expand_actions(action_descriptors, actions):
    result = []
    if not isinstance(actions, (list, tuple)):
        actions = [actions]
    for action in actions:
        if not isinstance(action, dict):
            action = {'id': action}
        action_id = action.pop('id')
        action_data = action
        result.append(action_descriptors[action_id].clone(**action_data))
    return result


class Application:

    def __init__(self, exception_handler=None):
        self._exception_handler = exception_handler
        self.sensors = {}
        self.states = {}
        self.action_descriptors = {}
        self.state = None
        self.logger = get_logger()
        self.loop = asyncio.get_event_loop()
        self._app_run_future = None
        self.config_path = None

    async def load_config(self, config_path,
                          default_state_factory=State,
                          default_action_descriptor_factory=ActionDescriptor,
                          default_schedule_factory=Schedule):
        with open(os.path.join(config_path, MAIN_CONFIG_FILE)) as config_file:
            config = yaml.load(config_file)
        self.config_path = config_path

        for action_id, action_data in config.get('actions', {}).items():
            abstract = action_data.pop('abstract', False)
            if not abstract:
                action_class = import_class(action_data.pop('class'))
                action = default_action_descriptor_factory(self, action_id, action_class, **action_data)
                self.add_action_descriptor(action_id, action)

        for sensor_id, sensor_data in config.get('sensors', {}).items():
            sensor_class = import_class(sensor_data.pop('class'))
            behaviours = sensor_data.pop('behaviours', {})
            sensor = sensor_class(self, sensor_id, **sensor_data)
            for code, actions in behaviours.items():
                for action_descriptor in expand_actions(self.action_descriptors, actions):
                    sensor.add_behaviour(code, action_descriptor)
            self.add_sensor(sensor_id, sensor)

        for state_id, state_data in config.get('states', {}).items():
            state_class = default_state_factory
            sensors = {}
            behaviours = []
            for sensor in state_data.pop('sensors', []):
                if isinstance(sensor, dict):
                    sensor_id = sensor['id']
                    for behaviour_code, actions in sensor['behaviours'].items():
                        for action_descriptor in expand_actions(self.action_descriptors, actions):
                            behaviours.append((sensor_id, behaviour_code, action_descriptor))
                else:
                    sensor_id = sensor
                sensors[sensor_id] = self.sensors[sensor_id]
            schedules_data = state_data.pop('schedules', {})
            state = state_class(self, state_id, sensors, **state_data)
            for behaviour in behaviours:
                state.add_behaviour(*behaviour)
            for schedule_id, schedule_data in schedules_data.items():
                schedule_class = default_schedule_factory
                actions = schedule_data.pop('actions')
                schedule = schedule_class(self, schedule_id, state, **schedule_data)
                for action_descriptor in expand_actions(self.action_descriptors, actions):
                    schedule.add_action(action_descriptor)
                state.add_schedule(schedule_id, schedule)
            self.add_state(state_id, state)

        await self.set_state(self.states[config['initial_state']])

    def exception_handler(self, exception):
        if self._exception_handler:
            return self._exception_handler(exception)
        raise exception

    def add_sensor(self, id_, sensor):
        self.sensors[id_] = sensor

    def add_state(self, id_, state):
        self.states[id_] = state

    def add_action_descriptor(self, id_, action_descriptor):
        self.action_descriptors[id_] = action_descriptor

    async def set_state(self, state):
        real = self.state != state or state.reactivatable
        self.logger.info('set_state', state=state.id, old_state=self.state.id if self.state else None, ignore=not real)
        if real:
            if self.state:
                await self.state.deactivate()
            self.state = state
            await self.state.activate()

    async def run(self):
        self.logger.info('application_start')
        sensor_tasks = [asyncio.ensure_future(sensor.run_forever())
                        for sensor in self.sensors.values()]
        self._app_run_future = asyncio.Future()
        await self._app_run_future
        self._app_run_future = None
        for sensor in self.sensors.values():
            sensor.stop()
        await asyncio.wait(sensor_tasks)
        if self.state:
            await self.state.deactivate()
        self.logger.info('application_end')

    def stop(self):
        self.logger.info('application_stop')
        if self._app_run_future:
            self._app_run_future.set_result(None)

    async def notify(self, sensor, code):
        if self.state:
            return await self.state.notify(sensor, code)
        self.logger.info('notify_ignore', reason='no_active_state')



import sys
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

import aiosmtplib
from jinja2 import Environment, PackageLoader, Template

from alarme import Action


class EmailAction(Action):

    def __init__(self, app, id_, host, port, login, password, sender, recipient, subject, env={}, text=None, template=None):
        super().__init__(app, id_)
        self.host = host
        self.port = port
        self.login = login
        self.password = password
        self.sender = sender
        self.recipient = recipient
        self.subject = subject
        self.env = env
        self.text = text
        self.template = template

    async def run(self):
        self.logger.info('send_email')
        smtp = aiosmtplib.SMTP(hostname=self.host, port=self.port)
        await smtp.connect()
        await smtp.ehlo()
        await smtp.starttls()
        await smtp.ehlo()
        await smtp.login(self.login, self.password)
        env = self.env.copy()
        env.update(dict(
            app=self.app,
        ))
        package_name = sys.modules[__name__].__package__
        jinja_env = Environment(loader=PackageLoader(package_name))
        try:
            plain_template = jinja_env.get_template(
                '{}.txt'.format(self.template)
            )
        except:
            plain_template = Template(self.text)
        part1 = MIMEText(plain_template.render(**env), 'plain')
        try:
            html_template = jinja_env.get_template(
                '{}.html'.format(self.template)
            )
        except:
            msg = part1
        else:
            part2 = MIMEText(html_template.render(**env), 'html')
            msg = MIMEMultipart('alternative')
            msg.attach(part1)
            msg.attach(part2)
        msg['Subject'] = Template(self.subject).render(**env)
        await smtp.sendmail(self.sender, [self.recipient], msg.as_string())



import asyncio

from .essential import Essential


class ScheduleStop(Exception):
    pass


class Schedule(Essential):

    def __init__(self, app, id_, state, run_count=None, run_interval=1, delay=0):
        super().__init__(app, id_)
        self.state = state
        self.run_count = run_count
        self.run_interval = run_interval
        self.delay = delay
        self.actions = []
        self.active_action = None
        self._future = None
        self._sleep_future = None
        self.running = False

    def add_action(self, action_descriptor):
        self.actions.append(action_descriptor)

    def _end(self):
        self.logger.info('schedule_end')
        self._future = None
        raise ScheduleStop

    def _end_sleep(self):
        if self._future and not self._future.done():
            self._future.set_result(None)

    async def _sleep(self, delay):
        # TODO: Too complicated, reimplement it
        if not self.running:
            self._end()
        self._future = asyncio.Future()
        self._sleep_future = self.loop.call_later(delay, self._end_sleep)
        try:
            await self._future
        except asyncio.CancelledError:
            self._end()

    def _continue(self, run_count):
        return (self.run_count is None or run_count < self.run_count) and self.running

    async def run_actions(self):
        for action_descriptor in self.actions:
            self.active_action = action_descriptor.construct()
            try:
                while self.running:
                    try:
                        await self.active_action.execute()
                    except:
                        await self._sleep(1)
                    else:
                        break
            finally:
                self.active_action = None
            if not self.running:
                break

    async def run(self):
        self.logger.info('schedule_run', delay=self.delay)
        self.running = True
        try:
            await self._sleep(self.delay)
            run_count = 0
            while self._continue(run_count):
                await self.run_actions()
                run_count += 1
                if self._continue(run_count):
                    self.logger.info('schedule_interval', interval=self.run_interval)
                    await self._sleep(self.run_interval)
            self._end()
        except ScheduleStop:
            pass
        finally:
            self.running = False

    def stop(self):
        self.logger.info('schedule_stop')
        if self._future:
            self._future.cancel()
        if self._sleep_future:
            self._sleep_future.cancel()
        if self.active_action:
            self.active_action.stop()
        self.running = False



from gsm_modem_asyncio import GsmModem

from alarme import Action


class SingleGsmModemMeta(type):

    instances = {}

    def __call__(mcs, serial_url, *args, **kwargs):
        if serial_url not in mcs.instances:
            mcs.instances[serial_url] = super().__call__(serial_url, *args, **kwargs)
        return mcs.instances[serial_url]


class SingleGsmModem(GsmModem, metaclass=SingleGsmModemMeta):
    pass


class GsmAction(Action):

    def __init__(self, app, id_, serial_url):
        super().__init__(app, id_)
        self.serial_url = serial_url
        self.gsm_modem = SingleGsmModem(self.serial_url)

    async def run(self):
        raise NotImplementedError



import asyncio

from alarme import Sensor
from alarme.extras.common import SingleRFDevice


class RfReceiverSensor(Sensor):

    def __init__(self, app, id_, gpio, code, check_interval=0.1):
        super().__init__(app, id_)
        self.gpio = gpio
        self.code = code
        self.check_interval = check_interval
        self.rf_device = SingleRFDevice(self.gpio)

    async def process_code(self, raw_code):
        event = raw_code ^ self.code
        if event in self.behaviours:
            await self.notify(event)

    async def cleanup(self):
        await super().cleanup()
        # self.rf_device.cleanup()

    async def run(self):
        self.rf_device.enable_rx()
        try:
            timestamp = self.rf_device.rx_code_timestamp
            while self.running:
                if self.rf_device.rx_code_timestamp != timestamp:
                    timestamp = self.rf_device.rx_code_timestamp
                    raw_code = self.rf_device.rx_code
                    self.logger.debug('checking_signal',
                                      raw_code=raw_code,
                                      pulselength=self.rf_device.rx_pulselength,
                                      protocol=self.rf_device.rx_proto)
                    await self.process_code(raw_code)
                await asyncio.sleep(self.check_interval)
        finally:
            self.rf_device.disable_rx()



import os.path
import sys
import logging.config

import structlog
from structlog import get_logger

from alarme.overrides.logging_extras import ProcessorFormatter, event_dict_to_message


def uncaught_exception(value):
    logger = get_logger()
    try:
        print(value)
        raise value
    except:
        logger.critical('uncaught_exception', name=type(value).__name__, args=value.args, exc_info=True)


def sys_uncaught_exception(exctype, value, tb):
    return uncaught_exception(value)


def loop_uncaught_exception(loop, exc_info):
    exception = exc_info.get('exception')
    if exception:
        return uncaught_exception(exception)


def init_logging(log_dir, name):
    if not os.path.exists(log_dir):
        os.mkdir(log_dir)
    plain_processor = structlog.dev.ConsoleRenderer() # TODO: ConsoleRenderer without coloring
    color_processor = structlog.dev.ConsoleRenderer()
    plain_processor._level_to_color['info'] = structlog.dev.BRIGHT
    color_processor._level_to_color['info'] = structlog.dev.BRIGHT
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'plain': {
                '()': ProcessorFormatter,
                'processor': plain_processor,
            },
            'colored': {
                '()': ProcessorFormatter,
                'processor': color_processor,
            },
        },
        'handlers': {
            'default': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'colored',
            },
            'file': {
                'level': 'DEBUG',
                'class': 'logging.handlers.WatchedFileHandler',
                'filename': os.path.join(log_dir, '{}.log'.format(name)),
                'formatter': 'plain',
            },
        },
        'loggers': {
            '': {
                'handlers': ['default', 'file'],
                'level': 'DEBUG',
                'propagate': True,
            },
            'asyncio': {
                'propagate': False,
            },
        }
    })
    structlog.configure(
        processors=[
            structlog.stdlib.add_log_level,
            structlog.stdlib.PositionalArgumentsFormatter(),
            structlog.processors.TimeStamper(fmt='%Y-%m-%d %H:%M:%S'),
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            event_dict_to_message,
        ],
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )
    sys.excepthook = sys_uncaught_exception



from aiohttp_jinja2 import template

from .core import CoreView, http_found
from ..util import login_required, handle_exception


class Control(CoreView):

    @handle_exception
    @login_required
    @template('control.html')
    async def get(self):
        self.logger.debug('control_page_view')
        return {'current_state': self.sensor.app.state,
                'buttons': self.sensor.buttons,
                'states': self.sensor.app.states}

    @handle_exception
    @login_required
    @http_found
    async def post(self):
        data = await self.request.post()
        behaviour = data['behaviour']
        futures = await self.sensor.notify(behaviour)
        for future in (futures or []):
            await future



"""
Solution of separating Colored/non-colored log by formatters/handlers:
https://github.com/hynek/structlog/pull/73#issuecomment-222931463
TODO: Remove when https://github.com/hynek/structlog/pull/79 will be released to pip
"""


import logging


class ProcessorFormatter(logging.Formatter):
    """Custom stdlib logging formatter for structlog ``event_dict`` messages.
    Apply a structlog processor to the ``event_dict`` passed as
    ``LogRecord.msg`` to convert it to loggable format (a string).
    """

    def __init__(self, processor, fmt=None, datefmt=None, style='%'):
        """Keep reference to the ``processor``."""
        super().__init__(fmt=fmt, datefmt=datefmt, style=style)
        self.processor = processor

    def format(self, record):
        """Extract structlog's ``event_dict`` from ``record.msg``.
        Process a copy of ``record.msg`` since the some processors modify the
        ``event_dict`` and the ``LogRecord`` will be used for multiple
        formatting runs.
        """
        if isinstance(record.msg, dict):
            msg_repr = self.processor(
                record._logger, record._name, record.msg.copy())
            return msg_repr
        return record.getMessage()


def event_dict_to_message(logger, name, event_dict):
    """Pass the event_dict to stdlib handler for special formatting."""
    return ((event_dict,), {'extra': {'_logger': logger, '_name': name}})



from .essential import Essential


class Action(Essential):

    def __init__(self, app, id_):
        super().__init__(app, id_)
        self.running = True

    async def execute(self):
        self.logger.info('action_run')
        try:
            result = await self.run()
        except:
            self.logger.error('action_crash', exc_info=True)
            raise
        else:
            self.logger.info('action_end')
        finally:
            await self.cleanup()
        return result

    async def run(self):
        pass

    def stop(self):
        self.logger.info('action_stop')
        self.running = False

    async def cleanup(self):
        self.logger.info('action_cleanup')



import asyncio
import os.path

import pyglet

from alarme import Action


class SoundAction(Action):

    def __init__(self, app, id_, sound):
        super().__init__(app, id_)
        self.sound = sound
        self._future = None

    async def run(self):
        self.logger.info('play_sound')
        player = pyglet.media.Player()
        source = pyglet.media.load(os.path.join(self.app.config_path or '', self.sound), streaming=False)
        # player.push_handlers(on_eos=self.end) # Not working, so do it using call_later() and source.duration
        end_task = self.loop.call_later(source.duration, self.end)
        player.queue(source)
        player.play()
        self._future = asyncio.Future()
        await self._future
        end_task.cancel()
        player.pause()

    def end(self):
        if self._future and not self._future.done():
            self._future.set_result(None)

    def stop(self):
        super().stop()
        self.end()



from warnings import warn

try:
    from rpi_rf import RFDevice
except RuntimeError as err:
    warn('\n\n'.join(err.args))
    RFDevice = None


class SingleRFDeviceMeta(type):

    instances = {}

    def __call__(mcs, gpio, *args, **kwargs):
        if gpio not in mcs.instances:
            mcs.instances[gpio] = super().__call__(gpio, *args, **kwargs)
        return mcs.instances[gpio]

if RFDevice:
    class SingleRFDevice(RFDevice, metaclass=SingleRFDeviceMeta):
        pass
else:
    from unittest.mock import MagicMock
    SingleRFDevice = MagicMock()



import alarme

from .core import CoreView
from ..util import login_required, handle_exception, json_response, expand_color


class Info(CoreView):

    @handle_exception
    @login_required
    @json_response
    async def get(self):
        states = self.sensor.app.states
        buttons = self.sensor.buttons.copy()
        for button in buttons.values():
            button['color'] = expand_color(button['color'], states)
        # TODO: add behaviours to response
        return {'app': {'version': alarme.__version__},
                'current_state': self.sensor.app.state.as_dict(),
                'buttons': buttons,
                'states': {state.id: state.as_dict()
                           for state in states.values()}}



from .gsm_action import GsmAction


class SmsAction(GsmAction):

    def __init__(self, app, id_, serial_url, number, text, timeout=20):
        super().__init__(app, id_, serial_url)
        self.number = number
        self.text = text
        self.timeout = timeout

    async def run(self):
        self.logger.info('send_sms')
        await self.gsm_modem.send_sms(self.number, self.text, self.timeout)



import asyncio

from structlog import get_logger


class Essential:

    def __init__(self, app, id_):
        self.app = app
        self.id = id_
        self.logger = get_logger(id=self.id)
        self.loop = asyncio.get_event_loop()



from aiohttp.web import View, HTTPFound


def http_found(func):
    async def wrapped(self, *args, **kwargs):
        await func(self, *args, **kwargs)
        return HTTPFound(self.request.rel_url)
    return wrapped


class CoreView(View):

    sensor = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = self.sensor.logger



import asyncio

from alarme import Action


class SetStateAction(Action):

    def __init__(self, app, id_, state_id):
        super().__init__(app, id_)
        self.state_id = state_id

    async def run(self):
        # Can't await cause this action must end in current state deactivation, so asyncio.ensure_future
        return asyncio.ensure_future(self.app.set_state(self.app.states[self.state_id]))



import asyncio

from .essential import Essential


class Sensor(Essential):

    def __init__(self, app, id_):
        super().__init__(app, id_)
        self.behaviours = {}
        self.running = True

    def add_behaviour(self, code, action_descriptor):
        self.behaviours.setdefault(code, []).append(action_descriptor)

    async def run_forever(self):
        while self.running:
            self.logger.info('sensor_run')
            try:
                await self.run()
            except:
                self.logger.error('sensor_crash', exc_info=True)
                await asyncio.sleep(1)
            else:
                self.logger.info('sensor_end')
        await self.cleanup()

    async def run(self):
        while self.running:
            await asyncio.sleep(0.1)

    def stop(self):
        self.logger.info('sensor_stop')
        self.running = False

    async def cleanup(self):
        self.logger.info('sensor_cleanup')

    async def notify(self, code):
        self.logger.info('sensor_notify', code=code)
        return await self.app.notify(self, code)



import re
import os
from setuptools import setup
from pip.req import parse_requirements


install_reqs = parse_requirements('requirements.txt', session=False)
requirements = [str(ir.req) for ir in install_reqs]
package_name = 'alarme'
hyphen_package_name = package_name.replace('_', '-')


def read_version():
    regexp = re.compile(r"^__version__\s*=\s*'([\d.abrc]+)'")
    init_py = os.path.join(os.path.dirname(__file__), package_name, '__init__.py')
    with open(init_py) as f:
        for line in f:
            match = regexp.match(line)
            if match is not None:
                return match.group(1)
        else:
            raise RuntimeError('Cannot find version in {}'.format(init_py))


setup(
    name=hyphen_package_name,
    version=read_version(),
    packages=[package_name],
    url='https://github.com/insolite/{}'.format(hyphen_package_name),
    author='Oleg Krasnikov',
    author_email='a.insolite@gmail.com',
    description='Universal alarm system',
    include_package_data=True,
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            '{} = {}.scripts.server:main'.format(hyphen_package_name, package_name),
        ],
    },
)



from .essential import Essential


class ActionDescriptor(Essential):

    def __init__(self, app, id_, factory, **kwargs):
        super().__init__(app, id_)
        self.factory = factory
        self.kwargs = kwargs

    def construct(self, **kwargs):
        final_kwargs = self.kwargs.copy()
        final_kwargs.update(kwargs)
        return self.factory(self.app, self.id, **final_kwargs)

    def clone(self, **kwargs):
        final_kwargs = self.kwargs.copy()
        final_kwargs.update(kwargs)
        return self.__class__(self.app, self.id, self.factory, **final_kwargs)



import asyncio

from .essential import Essential


class State(Essential):

    def __init__(self, app, id_, sensors, color='#ddd', reactivatable=True):
        super().__init__(app, id_)
        self.sensors = sensors
        self.color = color
        self.schedules = {}
        self.behaviours = {}
        self.reactivatable = reactivatable
        self._schedules_tasks = []
        self.running = False
        self.active_action = None

    def add_schedule(self, id_, schedule):
        self.schedules[id_] = schedule

    def add_behaviour(self, sensor_id, code, action_descriptor):
        self.behaviours.setdefault(sensor_id, {}).setdefault(code, []).append(action_descriptor)

    async def activate(self):
        self.logger.info('state_activate')
        self.running = True
        self._schedules_tasks = [asyncio.ensure_future(schedule.run())
                                 for schedule in self.schedules.values()]

    async def deactivate(self):
        self.logger.info('state_deactivate')
        self.running = False
        if self.active_action:
            self.active_action.stop()
        for schedule in self.schedules.values():
            schedule.stop()
        if self._schedules_tasks:
            await asyncio.wait(self._schedules_tasks)

    async def sensor_react(self, sensor, code, behaviour):
        logger = self.logger.bind(sensor=sensor.id, code=code)
        logger.info('sensor_react')
        result = []
        try:
            for action_descriptor in behaviour:
                self.active_action = action_descriptor.construct()
                try:
                    result.append(await self.active_action.execute())
                except:
                    pass  # TODO: Try again as in schedule?
                if not self.running:
                    break
        finally:
            self.active_action = None
        return result

    async def notify(self, sensor, code):
        logger = self.logger.bind(sensor=sensor.id, code=code)
        if sensor in self.sensors.values():
            behaviour = self.get_behaviour(sensor, code)
            if behaviour:
                return await self.sensor_react(sensor, code, behaviour)
            else:
                logger.error('notify_ignore', reason='unknown_behaviour')
        else:
            logger.info('notify_ignore', reason='sensor_not_listed_for_state')

    def get_behaviour(self, sensor, code):
        special_behaviour = self.behaviours.get(sensor.id, {}).get(code)
        return special_behaviour or sensor.behaviours.get(code)

    def as_dict(self):
        return dict(id=self.id,
                    color=self.color)



import os.path
import sys
from functools import partial

import aiohttp_jinja2
import jinja2
from aiohttp import web

from alarme import Sensor
from . import views, filters


def generate_core_view(sensor, original_handler):
    # Can't use lambda like this, getting AttributeError:
    # return lambda x: original_handler(x)
    # AttributeError: 'RequestHandler' object has no attribute '_middlewares'
    # Probably it is an aiohttp bug
    # TODO: Check it later / post bug report
    return type(original_handler.__name__, (original_handler,),
                {'sensor': sensor})


class WebSensor(Sensor):

    def __init__(self, app, id_, host, port, buttons=[], login=None, password=None, debug=False):
        super().__init__(app, id_)
        self.host = host
        self.port = port
        self.buttons = buttons
        self.login = login
        self.password = password
        generate_view = partial(generate_core_view, self)
        app = web.Application(debug=debug)
        package_name = sys.modules[__name__].__package__
        package_path, = sys.modules[package_name].__path__
        jinja_env = aiohttp_jinja2.setup(
            app, loader=jinja2.PackageLoader(package_name))
        filter_functions = [
            filters.expand_color,
        ]
        jinja_env.filters.update({filter_function.__name__: filter_function
                                  for filter_function in filter_functions})
        app.router.add_route('*', '/control', generate_view(views.Control), name='control')
        app.router.add_route('*', '/info', generate_view(views.Info), name='info')
        app.router.add_route('*', '/', generate_view(views.Home), name='home')
        app.router.add_static('/static', path=os.path.join(package_path, 'static'), name='static')
        self.web_handler = app.make_handler()
        self.serv = None
        self.web_app = app

    async def run(self):
        self.serv = await self.loop.create_server(self.web_handler, self.host, self.port)
        await super().run()

    async def cleanup(self):
        await super().cleanup()
        self.serv.close()
        if self.serv:
            await self.serv.wait_closed()
        await self.web_app.shutdown()
        await self.web_handler.finish_connections(3)
        await self.web_app.cleanup()
